import { defineComponent, h, PropType, provide, ref, watch } from 'vue'

import { CFormControlWrapper } from './../form/CFormControlWrapper'
import { CPicker } from '../picker'

import { CMultiSelectNativeSelect } from './CMultiSelectNativeSelect'
import { CMultiSelectOptions } from './CMultiSelectOptions'
import { CMultiSelectSelection } from './CMultiSelectSelection'

import { filterOptionsList, flattenArray, selectOptions } from './utils'
import type { Option, SelectedOption } from './types'

const CMultiSelect = defineComponent({
  name: 'CMultiSelect',
  props: {
    /**
     * Enables selection cleaner element.
     *
     * @default true
     */
    cleaner: {
      type: Boolean,
      default: true,
    },
    /**
     * Toggle the disabled state for the component.
     */
    disabled: {
      type: Boolean,
      default: false,
    },
    /**
     * Provide valuable, actionable feedback.
     *
     * @since 4.6.0
     */
    feedback: {
      type: String,
    },
    /**
     * Provide valuable, actionable feedback.
     *
     * @since 4.6.0
     */
    feedbackInvalid: {
      type: String,
    },
    /**
     * Provide valuable, actionable invalid feedback when using standard HTML form validation which applied two CSS pseudo-classes, `:invalid` and `:valid`.
     *
     * @since 4.6.0
     */
    feedbackValid: {
      type: String,
    },
    /**
     * The id global attribute defines an identifier (ID) that must be unique in the whole document.
     */
    id: {
      type: String,
    },
    /**
     * Set component validation state to invalid.
     *
     * @since 4.6.0
     */
    invalid: Boolean,
    /**
     * Add a caption for a component.
     *
     * @since 4.6.0
     */
    label: {
      type: String,
    },
    /**
     * It specifies that multiple options can be selected at once.
     *
     * @default true
     */
    multiple: {
      type: Boolean,
      default: true,
    },
    /**
     * List of option elements.
     */
    options: {
      type: Array as PropType<Option[]>,
      default: () => [],
    },
    /**
     * Sets maxHeight of options list.
     *
     * @default 'auto'
     */
    optionsMaxHeight: {
      type: [Number, String],
      default: 'auto',
    },
    /**
     * Sets option style.
     *
     * @values 'checkbox', 'text'
     * @default 'checkbox'
     */
    optionsStyle: {
      type: String,
      default: 'checkbox',
      validator: (value: string) => {
        return ['checkbox', 'text'].includes(value)
      },
    },
    /**
     * Specifies a short hint that is visible in the search input.
     *
     * @default 'Select...''
     */
    placeholder: {
      type: String,
      default: 'Select...',
    },
    /**
     * Enables search input element.
     */
    search: {
      type: [Boolean, String],
      default: true,
      validator: (value: boolean | string) => {
        if (typeof value == 'string') {
          return ['external'].includes(value)
        }
        if (typeof value == 'boolean') {
          return true
        }
        return false
      },
    },
    /**
     * Sets the label for no results when filtering.
     */
    searchNoResultsLabel: {
      type: String,
      default: 'no items',
    },
    /**
     * Enables select all button.
     *
     * @default true
     */
    selectAll: {
      type: Boolean,
      default: true,
    },
    /**
     * Sets the select all button label.
     *
     * @default 'Select all options'
     */
    selectAllLabel: {
      type: String,
      default: 'Select all options',
    },
    /**
     * Sets the selection style.
     *
     * @values 'counter', 'tags', 'text'
     * @default 'tags'
     */
    selectionType: {
      type: String,
      default: 'tags',
      validator: (value: string) => {
        return ['counter', 'tags', 'text'].includes(value)
      },
    },
    /**
     * Sets the counter selection label.
     *
     * @default 'item(s) selected'
     */
    selectionTypeCounterText: {
      type: String,
      default: 'item(s) selected',
    },
    /**
     * Size the component small or large.
     *
     * @values 'sm', 'lg'
     */
    size: {
      type: String,
      validator: (value: string) => {
        return ['sm', 'lg'].includes(value)
      },
    },
    /**
     * Add helper text to the component.
     *
     * @since 4.6.0
     */
    text: {
      type: String,
    },
    /**
     * Display validation feedback in a styled tooltip.
     *
     * @since 4.6.0
     */
    tooltipFeedback: Boolean,
    /**
     * Set component validation state to valid.
     *
     * @since 4.6.0
     */
    valid: Boolean,
    /**
     * Enable virtual scroller for the options list.
     *
     * @since 4.8.0
     */
    virtualScroller: Boolean,
    /**
     * Toggle the visibility of multi select dropdown.
     *
     * @default false
     */
    visible: {
      type: Boolean,
      default: false,
    },
    /**
     *
     * Amount of visible items when virtualScroller is set to `true`.
     *
     * @since 4.8.0
     */
    visibleItems: {
      type: Number,
      default: 10,
    },
  },
  emits: [
    /**
     * Execute a function when a user changes the selected option. [docs]
     */
    'change',
    /**
     * Execute a function when the filter value changed.
     *
     * @since 4.7.0
     */
    'filterChange',
  ],
  setup(props, { attrs, emit }) {
    const nativeSelectRef = ref<HTMLSelectElement>()
    const options = ref<Option[]>(props.options)
    const search = ref('')
    const searchRef = ref<HTMLInputElement>()
    const selected = ref<SelectedOption[]>([])
    const visible = ref<Boolean>(props.visible)

    provide('nativeSelectRef', nativeSelectRef)

    watch(
      () => props.options,
      (newValue, oldValue) => {
        if (JSON.stringify(newValue) !== JSON.stringify(oldValue)) {
          options.value = newValue
          const _selected =
            newValue &&
            flattenArray(newValue).filter((option: Option) => {
              if (option.selected) {
                return option
              }

              return
            })

          const deselected =
            newValue &&
            flattenArray(newValue).filter((option: Option) => {
              if (option.selected === false) {
                return option
              }

              return
            })

          if (_selected) {
            selected.value = selectOptions(_selected, selected.value, deselected)
          }
        }
      },
      { immediate: true },
    )

    watch(selected, () => {
      nativeSelectRef.value &&
        nativeSelectRef.value.dispatchEvent(new Event('change', { bubbles: true }))
    })

    const handleSearchChange = (event: InputEvent) => {
      const target = event.target as HTMLInputElement
      search.value = target.value.toLowerCase()

      emit('filterChange', target.value)
    }

    const handleSearchKeyDown = (event: KeyboardEvent) => {
      if (search.value.length) {
        return
      }

      if (event.key === 'Backspace' || event.key === 'Delete') {
        const last = selected.value.filter((option: Option) => !option.disabled).pop()
        if (last) {
          selected.value = selected.value.filter((option: Option) => option.value !== last.value)
        }
      }
    }

    const handleOptionClick = (option: Option) => {
      if (!props.multiple) {
        selected.value = [{ value: option.value, text: option.text }] as SelectedOption[]
        visible.value = false
        search.value = ''
        if (searchRef.value) {
          searchRef.value.value = ''
        }

        return
      }

      if (selected.value.some((_option) => _option.value === option.value)) {
        selected.value = selected.value.filter((_option) => _option.value !== option.value)
      } else {
        selected.value = [
          ...selected.value,
          { value: option.value, text: option.text },
        ] as SelectedOption[]
      }
    }

    const handleSelectAll = () => {
      selected.value = selectOptions(
        flattenArray(options.value).filter((option: Option) => !option.disabled),
        selected.value,
      )
    }

    const handleDeselectAll = () => {
      selected.value = selected.value.filter((option) => option.disabled)
    }

    return () => [
      h(CMultiSelectNativeSelect, {
        multiple: props.multiple,
        options: options.value,
        value: props.multiple
          ? selected.value.map((option: Option) => option.value.toString())
          : selected.value.map((option: Option) => option.value)[0],
        onChange: () => emit('change', selected.value),
      }),
      h(
        CFormControlWrapper,
        {
          describedby: attrs['aria-describedby'],
          feedback: props.feedback,
          feedbackInvalid: props.feedbackInvalid,
          feedbackValid: props.feedbackValid,
          id: props.id,
          invalid: props.invalid,
          label: props.label,
          text: props.text,
          tooltipFeedback: props.tooltipFeedback,
          valid: props.valid,
        },
        {
          default: () =>
            h(
              CPicker,
              {
                class: [
                  'form-multi-select',
                  {
                    'form-multi-select-with-cleaner': props.cleaner,
                    disabled: props.disabled,
                    [`form-multi-select-${props.size}`]: props.size,
                    'form-multi-select-selection-tags':
                      props.multiple && props.selectionType === 'tags',
                    show: visible.value,
                    'is-invalid': props.invalid,
                    'is-valid': props.valid,
                  },
                ],
                disabled: props.disabled,
                id: props.id,
                onHide: () => {
                  visible.value = false
                },
                onShow: () => {
                  props.search && searchRef.value && searchRef.value.focus()
                  visible.value = true
                },
                visible: visible.value,
              },
              {
                toggler: () =>
                  h('div', {}, [
                    h(CMultiSelectSelection, {
                      multiple: props.multiple,
                      onRemove: (option: Option) => !props.disabled && handleOptionClick(option),
                      search: props.search,
                      selected: selected.value,
                      selectionType: props.selectionType,
                      selectionTypeCounterText: props.selectionTypeCounterText,
                    }),
                    props.multiple &&
                      props.cleaner &&
                      selected.value.length > 0 &&
                      !props.disabled &&
                      h('button', {
                        type: 'button',
                        class: 'form-multi-select-selection-cleaner',
                        onClick: () => handleDeselectAll(),
                      }),
                    props.search &&
                      h('input', {
                        type: 'text',
                        class: 'form-multi-select-search',
                        autocomplete: 'off',
                        ...(selected.value.length === 0 && { placeholder: props.placeholder }),
                        ...(selected.value.length &&
                          props.selectionType === 'counter' && {
                            placeholder: `${selected.value.length} ${props.selectionTypeCounterText}`,
                          }),
                        ...(selected.value.length &&
                          !props.multiple && {
                            placeholder: selected.value.map((option) => option.text)[0],
                          }),
                        disabled: props.disabled,
                        onInput: (event: InputEvent) => handleSearchChange(event),
                        onKeydown: (event: KeyboardEvent) => handleSearchKeyDown(event),
                        ...(props.multiple &&
                          selected.value.length &&
                          props.selectionType !== 'counter' && { size: search.value.length + 2 }),
                        ref: searchRef,
                      }),
                  ]),
                default: () =>
                  visible.value && [
                    props.multiple &&
                      props.selectAll &&
                      h(
                        'button',
                        {
                          class: 'form-multi-select-all',
                          onClick: () => handleSelectAll(),
                          type: 'button',
                        },
                        props.selectAllLabel,
                      ),
                    h(CMultiSelectOptions, {
                      onOptionClick: (option: Option) => handleOptionClick(option),
                      options:
                        props.search === 'external'
                          ? options.value
                          : filterOptionsList(search.value, options.value),
                      optionsMaxHeight: props.optionsMaxHeight,
                      optionsStyle: props.optionsStyle,
                      searchNoResultsLabel: props.searchNoResultsLabel,
                      selected: selected.value,
                      virtualScroller: props.virtualScroller,
                      visibleItems: props.visibleItems,
                    }),
                  ],
              },
            ),
        },
      ),
    ]
  },
})

export { CMultiSelect }
